/****************************************************************************
 * Copyright (C) 2010  Michael Oberg                                        *
 *                                                                          *
 * This file is part of the AjaxSync framework.                             *
 *                                                                          *
 * This program is free software; you can redistribute it and/or modify     *
 * it under the terms of the GNU General Public License as published by     *
 * the Free Software Foundation; either version 2 of the License, or        *
 * (at your option) any later version.                                      *
 *                                                                          *
 * This program is distributed in the hope that it will be useful,          *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 * GNU General Public License for more details.                             *
 *                                                                          *
 * You should have received a copy of the GNU General Public License        *
 * along with this program; if not, write to the Free Software              *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                *
 ****************************************************************************/

/**
 * \file synchronizer.cpp
 * \author Michael Oberg
 *
 * This is the main file of the AjaxSync Synchronizer service.
 *
 * This file will for now only compile under Linux. The most incompatible part
 * is the libuuid, which is part of the e2fstools package (ext2 and higher
 * filesystem) for Linux, but this library could in theory be extracted and
 * ported to other systems.
 * 
 * The next incompatible parts are the pthreads and socket libraries, which are
 * essentially only available for POSIX compatible UNIX systems (especially not
 * Microsoft Windows). The fork call does not exists in Windows, either - a
 * Windows service is send in the background by the starter, it does not detach
 * itself from the terminal.
 */

// C
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <unistd.h>
#include <uuid/uuid.h>
#include <pthread.h>
#include <string.h>
#include <sys/stat.h>
// C++
#include <string>
#include <map>

/**
 * Each contact with the synchronizer has to be authenticated with this password.
 * Max length is 32 characters (the 33 is for the terminating zero byte).
 */
char password[33];
/**
 * Log level is 0 for error, 1 for warnings, 2 for info/debugging
 */
int logLevel = 2;
/**
 * Pause between two keepalive signals in seconds
 */
int delay;

FILE *logfile;

typedef std::map<std::string, int> strlist_t; /**< only needed for signal_map, session_map */

/**
 * For each signal a map of sessions, which have registered listeners for this signal, is hold
 */
std::map<std::string, strlist_t> signal_map;

/**
 * For each session a map of signals, for which the session has registered listeners, is hold
 */
std::map<std::string, strlist_t> session_map;

/**
 * For each session this buffers one signal which has to be send to the session
 */
std::map<std::string, std::string> sig_buffer;

/**
 * For each session this stores the wakeup condition when the session is waiting for a signal
 */
std::map<std::string, pthread_cond_t*> sig_cond;

/**
 * Mutex for accessing the signal/session map, the signal buffer or condition map
 */
pthread_mutex_t sig_mutex = PTHREAD_MUTEX_INITIALIZER;

/**
 * Mutex for accessing the log file
 */
pthread_mutex_t log_mutex = PTHREAD_MUTEX_INITIALIZER;

/**
 * Splits a string into lines, suppresses carriage returns, and determines the lines length
 * @param s Input string
 * @param l List of output strings, each holding a line
 * @param slen Length of the input string (is not detected by zero byte)
 * @param llen Input: maximum number of lines to return; Output: number of lines returned
 */
void splitString(char *s, char **l, int slen, int *llen) {
    int i = 0;
    int j = 1;
    l[0] = s;
    while (i < slen && s[i] != 0 && j < *llen) {
        if (s[i] == '\n') {
            s[i] = 0;
            l[j++] = &(s[i + 1]);
        } else if (s[i] == '\r')
            s[i] = 0;
        i++;
    }
    *llen = j;
}

/**
 * Fills a string with zero bytes (needed for initializing records in socket operations)
 * @param s Input/Output string
 * @param n String length
 */
void setNull(char *s, int n) {
    for (int i = 0; i < n; i++)
        s[i] = 0;
}

/**
 * Write a log entry into the file (or to the standard output)
 * @param e The entry to write
 * @param l Level of the entry (error/warning/debug)
 */
void logEntry(const char* e, int l) {
    time_t rawtime;
    time(&rawtime);
    tm* t = localtime(&rawtime);
    pthread_mutex_lock(&log_mutex);
    if (l <= logLevel) {
        fprintf(logfile, "%04d-%02d-%02d %02d:%02d:%02d %s\n",
                t->tm_year + 1900, t->tm_mon + 1, t->tm_mday, t->tm_hour, t->tm_min, t->tm_sec, e);
        fflush(logfile);
    }
    pthread_mutex_unlock(&log_mutex);
}

/**
 * Write a log entry consisting of two strings in two lines; the second line starts with a prompt ">"
 * @param e First entry
 * @param p Second entry
 * @param l Level of the entry (error/warning/debug)
 */
void logEntry2(const char* e, const char* p, int l) {
    char buffer[1024];
    sprintf(buffer, "%s\n> %s", e, p);
    logEntry(buffer, l);
}

/**
 * Write a log entry consisting of three strings in three lines; the second and third line starts with a prompt ">"
 * @param e First entry
 * @param p1 Second entry
 * @param p2 Third entry
 * @param l Level of the entry (error/warning/debug)
 */
void logEntry3(const char* e, const char* p1, const char* p2, int l) {
    char buffer[1024];
    sprintf(buffer, "%s\n> %s\n> %s", e, p1, p2);
    logEntry(buffer, l);
}

/**
 * Check if the given session is still valid; unlock the mutex and stop the
 * thread if the session is expired. A locked mutex is required!
 * @param sessionId Pointer to the session id string
 * @param sockfd Socket for sending the answer in case of expiration
 */
void checkSession(char* sessionId, long int sockfd) {
    if (!session_map[sessionId]["keepalive"]) {
        pthread_mutex_unlock(&sig_mutex);
        logEntry2("Session has expired", sessionId, 1); // log warning
        shutdown(sockfd, SHUT_RD);
        const char *expireStr = "Session has expired\n";
        int n = send(sockfd, expireStr, strlen(expireStr), 0); // return warning
        if (n == EPIPE)
            logEntry("Broken pipe", 0);
        shutdown(sockfd, SHUT_RDWR);
        close(sockfd);
        pthread_exit(NULL); // exit immediately
    }
}

/**
 * Main (worker) thread. Called with every connection attempt to the service.
 *
 * Every Synchronizer command read from the socket cli_sock consists of two, three
 * or four lines ended with a line feed or carriage return / line feed. The first
 * line contains a single upper case character which encodes the command; the
 * second line contains the authentication #password. The optional third line
 * contains the session id of the sender; only the start (S) command has no third
 * line because it returns/defines the session id. The optional fourth line contains
 * a signal name - used by registering/unregistering listeners (L/R) for that signal,
 * or for the kill command (K), which sends that signal.
 *
 * The Synchronizer sends an answer to the command in the case of a session
 * expiration, in case of the start command (S - the answer is the generated
 * session id) and in case of the wait command (W - the answer is the received
 * signal).
 *
 * Threads receiving kill (K) and wait (W) commands are communicating with each
 * other. A kill thread sends a wakeup command to each existing and blocked wait
 * thread whose session has a listener registered for the sent signal.
 *
 * Another thread #keepalive is sending wakeup commands to threads receiving wait (W)
 * commands and forces them to send a keepalive signal each #delay seconds.
 *
 * @param cli_sock The TCP socket from which the command has to be read
 * @return Nothing is returned. The thread was detached immediately after calling
 * and simply vanishes with pthread_exit(NULL).
 */
void *worker(void *cli_sock) {
    std::map<std::string, int>::iterator it;
    std::string s;
    long int sockfd; // in 64 bit systems, pointers are equivalent to long int, not int; sockfd it cutted down to int when calling recv/send
    char buffer[513]; // max 512, one byte reserved for string terminating zero
    // copy client socket
    logEntry("reading socket", 2);
    sockfd = (long int) cli_sock;
    // read command from socket
    int n = recv(sockfd, buffer, 512, 0);
    buffer[n] = 0; // terminate returned string; length is 512 max
    logEntry2("characters read", buffer, 2); // log the whole input in debug mode
    const char *validStr = "Session ID is valid\n";
    char *list[10];
    int l = 10;
    // parse command
    splitString(buffer, list, n, &l); // split the input string in 10 lines max
    if ((l < 2) || strcmp(list[0], password) || (strlen(list[1]) != 1)) {
        // at least 2 arguments are required; first argument is the
        // password and has to be correct; second has to be an one
        // character command; otherwise shut down connection
        logEntry("No or wrong password / command", 1); // log warning
        shutdown(sockfd, SHUT_RDWR);
        close(sockfd);
        pthread_exit(NULL); // exit immediately
    }
    if (list[1][0] != 'S') {
        if (l < 3) {
            // for every command except [S]tart session 3 arguments
            // are required; third argument is the session id
            logEntry2("No session id for command", list[1], 1); // log warning
            shutdown(sockfd, SHUT_RDWR);
            close(sockfd);
            pthread_exit(NULL); // exit immediately
        }
    }
    switch (list[1][0]) {
        case 'S': // [S]tart session
            shutdown(sockfd, SHUT_RD);
            // generate session id
            uuid_t uuid_bin;
            char uuid_hex[38];
            char uuid_hex_ln[38];
            uuid_generate(uuid_bin);
            uuid_unparse(uuid_bin, uuid_hex);
            strcpy(uuid_hex_ln, uuid_hex);
            strcat(uuid_hex_ln, "\n");
            n = send(sockfd, uuid_hex_ln, strlen(uuid_hex_ln), 0);
            if (n == EPIPE)
                logEntry("Broken pipe", 0);
            else if (n >= strlen(uuid_hex)) { // if the client has received the session id, register it
                logEntry2("S: creating session id", uuid_hex, 2);
                pthread_mutex_lock(&sig_mutex);
                // register listener for the keepalive signal
                // and create the signal map on the fly
                signal_map["keepalive"][uuid_hex] = 1;
                session_map[uuid_hex]["keepalive"] = 1;
                // create signal buffer
                sig_buffer[uuid_hex] = "";
                // create wakeup listener thread condition
                sig_cond[uuid_hex] = new pthread_cond_t;
                pthread_cond_init(sig_cond[uuid_hex], NULL);
                pthread_mutex_unlock(&sig_mutex);
            } else
                logEntry("S: session id not completely read", 1); // log warning
            break;
        case 'T': // [T]est session
            pthread_mutex_lock(&sig_mutex);
            checkSession(list[2],sockfd);
            pthread_mutex_unlock(&sig_mutex);
            // if the session id was invalid, an error message has already been sent
            logEntry2("T: testing session ID", list[2], 2); // debug mode: log this access when the session is valid
            n = send(sockfd, validStr, strlen(validStr), 0);
            if (n == EPIPE)
                logEntry("Broken pipe", 0);
            break;
        case 'C': // [C]lose session
            pthread_mutex_lock(&sig_mutex);
            checkSession(list[2],sockfd);
            logEntry2("C: closing session", list[2], 2); // debug mode: log this access
            // delete all registered listeners for this session
            for (it = session_map[list[2]].begin(); it != session_map[list[2]].end(); it++) {
                signal_map[it->first].erase(list[2]);
                session_map[list[2]].erase(it->first);
            }
            // delete listener map of this session
            session_map.erase(list[2]);
            // delete signal buffer of this session
            sig_buffer.erase(list[2]);
            // delete wakeup listener thread condition of this session
            delete sig_cond[list[2]];
            sig_cond.erase(list[2]);
            pthread_mutex_unlock(&sig_mutex);
            break;
        case 'L': // [L]istener registration
            if ((l >= 4) && strlen(list[3])) {
                logEntry3("L: register signal listener for session", list[3], list[2], 2); // debug mode
                pthread_mutex_lock(&sig_mutex);
                checkSession(list[2],sockfd);
                signal_map[list[3]][list[2]] = 1;
                session_map[list[2]][list[3]] = 1;
                pthread_mutex_unlock(&sig_mutex);
            } else
                logEntry2("L: no signal given", list[2], 1);
            break;
        case 'R': // [R]emove listener
            if ((l >= 4) && strlen(list[3])) {
                logEntry3("R: remove signal listener for session", list[3], list[2], 2); // debug mode
                if (strcmp(list[3], "keepalive") == 0)
                    logEntry2("R: error, trying to remove keepalive - blocked", list[2], 1); // log warning
                else {
                    pthread_mutex_lock(&sig_mutex);
                    checkSession(list[2],sockfd);
                    signal_map[list[3]].erase(list[2]);
                    session_map[list[2]].erase(list[3]);
                    pthread_mutex_unlock(&sig_mutex);
                }
            } else
                logEntry2("R: no signal given", list[2], 1); // log warning
            break;
        case 'K': // [K]ill (UNIX style signalling :-))
            if ((l >= 4) && strlen(list[3])) {
                logEntry3("K: send signal from session", list[3], list[2], 2); // debug mode
                pthread_mutex_lock(&sig_mutex);
                checkSession(list[2],sockfd);
                // for every registered listener (session id) ...
                for (it = signal_map[list[3]].begin(); it != signal_map[list[3]].end(); it++) {
                    // if a client is already waiting, send wakeup signal
                    if (sig_buffer[it->first].compare("wait") == 0) {
                        sig_buffer[it->first] = list[3];
                        pthread_cond_signal(sig_cond[it->first]);
                    } else // put the signal in the buffer
                        sig_buffer[it->first] = list[3];
                }
                pthread_mutex_unlock(&sig_mutex);
            } else
                logEntry2("K: no signal given", list[2], 1);
            break;
        case 'W': // [W]ait for a signal
            logEntry2("W: session waiting for signal", list[2], 2); // debug mode: log with session id
            shutdown(sockfd, SHUT_RD);
            pthread_mutex_lock(&sig_mutex);
            checkSession(list[2],sockfd);
            if (!sig_buffer[list[2]].length()) {
                // if no signal is in the buffer, wait for the wakeup
                sig_buffer[list[2]] = "wait";
                pthread_cond_wait(sig_cond[list[2]], &sig_mutex);
            }
            s = sig_buffer[list[2]] + '\n';
            sig_buffer[list[2]] = "";
            logEntry3("W: returning signal to session", s.c_str(), list[2], 2); // debug mode: log the returned signal
            n = send(sockfd, s.c_str(), s.length(), 0);
            if (n == EPIPE)
                logEntry("Broken pipe", 0);
            pthread_mutex_unlock(&sig_mutex);
            break;
        default: // unknown command, do nothing
            logEntry2("unknown command", list[1], 1); // log warning
            break;
    }
    shutdown(sockfd, SHUT_RDWR);
    close(sockfd);
    logEntry("stopping thread", 2); // for debugging, log if this position could be reached
    pthread_exit(NULL);
}

/**
 * This thread awakens all "wait" #worker threads every #delay seconds and forces
 * them to send a keepalive signal. For sessions which have no running "wait"
 * thread the keepalive is simply stored in the sessions signal buffer #sig_buffer.
 *
 * It also detects session expiration. If a keepalive signal is still in the buffer
 * of a session, then the communication is broken since #delay seconds, and therefore
 * the session is expired. This thread removes then all registered listeners from
 * this session in both #signal_map and #session_map and deletes the session entries
 * in #session_map, #sig_buffer and #sig_cond.
 * @return Nothing. This thread never ends.
 */
void *keepalive(void *) {
    std::map<std::string, int>::iterator it;
    std::map<std::string, int>::iterator it2;
    std::string itStr;
    logEntry("keepalive thread just started", 0);
    for (;;) {
        // every <delay> seconds ...
        sleep(delay);
        pthread_mutex_lock(&sig_mutex);
        // for every registered listener (session id) ...
        for (it = signal_map["keepalive"].begin(); it != signal_map["keepalive"].end(); it++) {
            // if the last keepalive is still in the buffer,
            // drop the session
            if (sig_buffer[it->first].compare("keepalive") == 0) {
                itStr = it->first;
                logEntry2("keepalive: remove session", itStr.c_str(), 2); // debug mode: log session id
                // delete all registered listeners for this session
                for (it2 = session_map[itStr].begin(); it2 != session_map[itStr].end(); it2++) {
                    signal_map[it2->first].erase(itStr);
                    session_map[itStr].erase(it2->first);
                }
                // delete listener map of this session
                session_map.erase(itStr);
                // delete signal buffer of this session
                sig_buffer.erase(itStr);
                // delete wakeup listener thread condition of this session
                delete sig_cond[itStr];
                sig_cond.erase(itStr);
            } else if (sig_buffer[it->first].compare("wait") == 0) {
                // if a client is already waiting,
                // send wakeup signal
                sig_buffer[it->first] = "keepalive";
                pthread_cond_signal(sig_cond[it->first]);
            } else // put the signal in the buffer
                sig_buffer[it->first] = "keepalive";
        }
        pthread_mutex_unlock(&sig_mutex);
    }
}


/**
 * Start routine. It asks for TCP port, #delay, #logLevel, #password and name of
 * the log file from standard input without prompting, then forks into background
 * and detaches from terminal. It opens the TCP port and starts the keepalive
 * thread and the main event loop, which itself starts a #worker thread for each
 * connection attempt.
 * @param argc Ignored
 * @param argv Ignored
 * @return Nothing. The program does not terminate normally, it has to be killed.
 */
int main(int argc, char** argv) {
    int port,len;
    pid_t pid;
    char logfilename[256];
    scanf("%d\n%d\n%d\n",&port,&delay,&logLevel);
    fgets(password,32,stdin);
    fgets(logfilename,255,stdin);
    if ((len = strlen(password)) < 2) {
        printf("Expecting password\n");
        exit(0);
    }
    password[len-1] = 0; // remove newline
    if ((len = strlen(logfilename)) < 2) {
        printf("Expecting logfile\n");
        exit(0);
    }
    logfilename[len-1] = 0; // remove newline

    setpgrp(); // Detach from control terminal. If this is not done, f.e. starting the
    // synchronizer from Apache (PHP) will attach it to Apaches process group; because
    // Apache is process group leader, its shutdown would kill the synchronizer.

    if (pid = fork()) {
        // Some strange effect will happen if the first print command is executed before
        // the fork: because printf first writes in a buffer, not immediately to stdout,
        // fork may duplicate this buffer, and it will be printed in both parent and child
        // process - in the parent, the buffer is emptied with exit(0), in the child with
        // close(stdout).
        printf("port=%d,delay=%d,loglevel=%d,logfile=%s\n",port,delay,logLevel,logfilename);
        printf("PID: %d\n",pid);
        exit(0);
    }
    // detaching from terminal; all output goes to the logfile
    logfile = fopen(logfilename,"a");
    fclose(stdin);
    fclose(stdout);
    fclose(stderr);

    // open the service TCP socket
    int srv_sock;
    srv_sock = socket(AF_INET, SOCK_STREAM, 0);
    if (srv_sock == -1) {
        fprintf(logfile, "Could not create socket\n");
        exit(1);
    }
    struct sockaddr_in srv_addr;
    setNull((char *) & srv_addr, sizeof (sockaddr_in));
    srv_addr.sin_family = AF_INET;
    srv_addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
    srv_addr.sin_port = htons(port);
    if (bind(srv_sock, (struct sockaddr *) & srv_addr, sizeof (srv_addr)) < 0) {
        fprintf(logfile, "Could not bind socket\n");
        exit(1);
    }
    listen(srv_sock, 5);

    // from now on, ignore broken pipes (connections) on writing
    // - otherwise the server will break down in such a case!
    (void) signal(SIGPIPE, SIG_IGN);

    // start the keepalive thread and detach it
    pthread_t t;
    pthread_create(&t, NULL, &keepalive, NULL);
    pthread_detach(t);

    // for each connection attempt, start a worker thread and detach it
    for (;;) {
        socklen_t cli_len = sizeof (sockaddr_in);
        int cli_sock_tmp = accept(srv_sock, (struct sockaddr *) & srv_addr, &cli_len);
        pthread_create(&t, NULL, &worker, (void*) cli_sock_tmp);
        pthread_detach(t);
    }

    // this is never reached
    return (EXIT_SUCCESS);
}
