<?php
/****************************************************************************
 * Copyright (C) 2010  Michael Oberg
 * 
 * This file is part of the AjaxSync framework.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ****************************************************************************/

/**
 * This class manages the delivery of function calls in AjaxSync, including the
 * buffers / queues.
 *
 * It is a container for static functions.
 *
 * Every method requires an opened ajaxsyncSession object (i.e. call to
 * #ajaxsyncSession::session_start() before and #ajaxsyncSession::session_write_close()
 * after calling the methods)
 *
 * @author Michael Oberg
 */
final class ajaxsyncBuffer {
    /**
     * A private constructor/clone does make creating instances of this class impossible.
     */
    private function __construct() { }

    /**
     * A private constructor/clone does make creating instances of this class impossible.
     */
    private function __clone() { }

    /**
     * Whenever a function handler is registered during the login, a queue has to
     * be created which holds the calls of this function till they can be delivered
     * to the client.
     * @param $fname Name of the function.
     * @return True, if the queue could be created - that means especially that read
     * permission (receiving calls of this function) is granted for the user. False,
     * if either permission is denied or the function does not exist.
     */
    public static function register($fname) {
        $session = ajaxsyncSession::getSID(); // now existence of the session is guaranteed
        $uid = ajaxsyncSession::$values["AJAXSYNC_UID"];
        if (!$uid) {
            throw new Exception("Failed: not logged in yet");
        }
        $fname = mysql_escape_string($fname);
        $id = ajaxsyncDB::insertId("insert ignore into queues (session, fid) select $session, f.id from functions f, groupmembers m where f.name = '$fname' and (f.owner = $uid and f.uread = 1 or m.user = $uid and m.group = f.group and f.gread = 1)");
        if ($id) { // queue could be created, so read permission is given for that user
            return true;
        }
        else { // either permission is denied, or function does not exist
            return false;
        }
    }

    /**
     * During logout or after an expiration of the login, every existing function handler
     * has to be unregistered; that means especially deleting all assiciated queues.
     */
    public static function unregisterAll() {
        $session = ajaxsyncSession::getSID(); // now existence of the session is guaranteed
        ajaxsyncDB::change("delete from callreferences using callreferences, queues q where callreferences.qid = q.id and q.session = $session");
        ajaxsyncDB::change("delete from queues where session = $session");
    }

    /**
     * When a function is deleted, every existing queue holding calls of this function
     * has to be emptied and deleted.
     * @param $fname Name of the function.
     */
    public static function deleteQueues($fname) {
        if ($fname = mysql_escape_string($fname)) {
            ajaxsyncDB::change("delete from callreferences using callreferences, queues q, functions f where callreferences.qid = q.id and q.fid = f.id and f.name = '$fname'");
            ajaxsyncDB::change("delete from queues using queues, functions f where queues.fid = f.id and f.name = '$fname'");
        }
    }

    /**
     * When an AjaxSync function is called, the functions name and its parameter are
     * stored in a buffer, provided the fact that write permission for this function
     * is given to the user.
     * @param $cid ID / number of the call given by the sender; this number is only
     * unique in the senders session.
     * @param $fname Name of the function to be called.
     * @param $param Parameter of the call. Basically a BLOB (Binary Large Object).
     * @param $timeout Server timeout in seconds. If the call cannot be delivered in
     * that many seconds after the server received the call, it is simply deleted.
     * @return True, if the call either could be made - that means especially that
     * write permission (calling this function) is granted for that user - or the call
     * was already made. False if write permission is denied.
     */
    public static function makeCall($cid, $fname, $param, $timeout) {
        $session = ajaxsyncSession::getSID(); // now existence of the session is guaranteed
        $uid = ajaxsyncSession::$values["AJAXSYNC_UID"];
        if (!$uid) {
            throw new Exception("Failed: not logged in yet");
        }
        if ($cid <= ajaxsyncSession::$values["AJAXSYNC_CID"]) {
            return true; // Call was already processed
        }
        ajaxsyncSession::$values["AJAXSYNC_CID"] = $cid;
        $param = mysql_escape_string($param);
        $received = time(); // UNIX timestamp as integer
        $callends = $received + $timeout;
        $id = ajaxsyncDB::insertId("insert into calls (fid, param, sender, cid, timeout, received, callends) select distinct f.id, '$param', $session, $cid, $timeout, $received, $callends from functions f, groupmembers m where f.name = '$fname' and m.user = $uid and (f.owner = m.user and f.uwrite = 1 or f.group = m.group and f.gwrite = 1)");
        if ($id) { // success - now insert the call in the receiver queues - that is each queue referencing this function except the queue of the sender
            ajaxsyncDB::change("insert into callreferences (`call`, qid) select c.id, q.id from queues q, calls c where c.id = $id and q.fid = c.fid and q.session != $session");
            return true; // It is irrelevant if there is any listener; if the insertion into "calls" was successful, return true
        }
        else {
            return false; // Permission denied or function doesn't exist
        }
    }

    /**
     * Get all calls for this session from its registered queues.
     * @return List of calls including the servers ID / number for them. The number is
     * used by the client for detecting double delivery.
     */
    public static function getCalls() {
        $session = ajaxsyncSession::getSID(); // now existence of the session is guaranteed
        $clientReceived = time(); // UNIX timestamp as integer
        ajaxsyncDB::change("delete from callreferences using callreferences, calls c where callreferences.call = c.id and c.callends < $clientReceived"); // garbage collection
        ajaxsyncDB::change("delete from calls where callends < $clientReceived"); // garbage collection
        return ajaxsyncDB::selectMultiple("select c.id, f.name, c.param from callreferences r, queues q, calls c, functions f where q.session = $session and r.qid = q.id and r.call = c.id and c.fid = f.id");
    }

    /**
     * Every function call received by the client is confirmed with its next wait call.
     * The confirmed calls have to be deleted from the queues in order to prevent
     * double delivery.
     * @param $cid Servers ID / number for this call.
     */
    public static function confirmCall($cid) {
        $session = ajaxsyncSession::getSID(); // now existence of the session is guaranteed
        ajaxsyncDB::change("delete from callreferences using callreferences, queues q where callreferences.call = $cid and callreferences.qid = q.id and q.session = $session");
    }
}
?>