<?php
/****************************************************************************
 * Copyright (C) 2010  Michael Oberg
 * 
 * This file is part of the AjaxSync framework.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ****************************************************************************/

/**
 * This is the interface between the Webserver and the Synchronizer service (synchronizer.cpp).
 * It is a container for static functions.
 *
 * This class references the #ajaxsyncSession class.
 *
 * @author Michael Oberg
 */
final class ajaxsyncCore {
    /**
     * A private constructor/clone does make creating instances of this class impossible.
     */
    private function __construct() { }

    /**
     * A private constructor/clone does make creating instances of this class impossible.
     */
    private function __clone() { }

    /**
     * The return values of most of the ajaxsyncCore functions are used for error
     * detection. If the value is empty ("false"), the call was successful. In this
     * case the buffer variable $answer contains the "real" return value.
     */
    private static $answer = '';

    /**
     * This buffer variable can be read with the #getReceivedSignal function.
     */
    private static $lastReceivedSignal = '';

    /**
     * Every call to the Synchronizer service is done using a TCP socket. This function
     * creates a TCP socket connecting to the Synchronizer at port #AJAXSYNC_PORT,
     * sends a string, optionally receives a string, and closes the socket.
     *
     * This is called only by #request and #requestReconnect.
     * @param $in Request sent to the Synchronizer service
     * @return The answer of the Synchronizer service
     */
    private static function ajaxsyncConnect($in) {
        // connect to the AjaxSync server on localhost
        $socket = socket_create(AF_INET, SOCK_STREAM, SOL_TCP);
        if (!$socket) {
            ajaxsyncCtl::startSynchronizer(); // no connection - try to (re-)start ...
            return "Failed: " . socket_strerror(socket_last_error()); // ... but show error anyway
        }
        $result = socket_connect($socket, '127.0.0.1', AJAXSYNC_PORT);
        if (!$result) {
            ajaxsyncCtl::startSynchronizer(); // no connection - try to (re-)start ...
            return "Failed: " . socket_strerror(socket_last_error($socket)); // ... but show error anyway
        }
        socket_write($socket, $in, strlen($in));
        socket_shutdown($socket, 1); // make the socket read-only; this is equivalent to sending an EOF to the Synchronizer, which in turn starts the processing of the request
        $out = socket_read($socket, 2048); // read the answer, waiting for similar EOF from the Synchronizer ...
        socket_close($socket); // ... and shut down completely (warning: TCP sockets are a very limited ressource!)
        return preg_replace("/\\n/","",$out);
    }

    /**
     * The commands of the Synchronizer can be parted in two groups: starting, closing
     * and testing a session does not require a valid Synchronizer session; every other
     * command requires a valid session. The #requestReconnect enforces the validity
     * in form of reestablishing a session if the session was broken; the #request
     * will simply fail if the session was broken.
     *
     * The session ID is taken from the #ajaxsyncSession variable AJAXSYNC_UUID.
     * @param $in Request sent to the Synchronizer service
     * @return The status of the Synchronizer session. The optional answer of the
     * Synchronizer is returned in #$answer.
     */
    private static function request($in) {
        $out = self::ajaxsyncConnect($in);
        if (preg_match("/Session has expired/",$out)) {
            ajaxsyncSession::session_start();
            unset(ajaxsyncSession::$values["AJAXSYNC_UUID"]);
            ajaxsyncSession::session_write_close();
            return $out;
        }
        else if (preg_match("/Session ID is valid|Failed:/",$out)) {
            return $out;
        }
        self::$answer = $out;
        return ""; // ok, answer is stored (only other answers are uuid's and signal names)
    }

    /**
     * The commands of the Synchronizer can be parted in two groups: starting, closing
     * and testing a session do not require a valid Synchronizer session; every other
     * command requires a valid session. The #requestReconnect enforces the validity
     * in form of reestablishing a session if the session was broken; the #request
     * will simply fail if the session was broken.
     *
     * Different from #request, the session ID is not taken from the #ajaxsyncSession
     * variable AJAXSYNC_UUID, instead it is already part of the parameter $in. But if
     * the session was not valid, the session id in the parameter $in has to be replaced;
     * only for this reason the session variable AJAXSYNC_UUID is then read and buffered.
     *
     * If the session was not valid at the first try, a new session is started with
     * #startSession. If that was successful, the #ajaxsyncSession variable AJAXSYNC_UUID
     * is updated with the new session ID, and the session id in the parameter $in is
     * replaced with the new value.
     *
     * Additionally every listener which was registered in the now invalid session (and
     * stored in the #ajaxsyncSession list variable AJAXSYNC_LISTENERS) is re-registered
     * using #listenerRegistration. If an error occures during these steps, it is returned.
     * @param $in Request sent to the Synchronizer service
     * @return Error message, if the session could not be (re-)established, or empty if
     * no error occures. The optional answer of the Synchronizer is returned in #$answer.
     */
    private static function requestReconnect($in) {
        $out = self::ajaxsyncConnect($in);
        if (preg_match("/Session has expired/",$out)) { // then start a new session and re-register all listeners!
            ajaxsyncSession::session_start();
            $old_uuid = ajaxsyncSession::$values["AJAXSYNC_UUID"];
            ajaxsyncSession::session_write_close();
            if ($errmsg = self::startSession()) {
                return $errmsg;
            } // stop in case of connection problems, do the same in the next steps ...
            ajaxsyncSession::session_start(); // this re-registers listeners, if possible ...
            $new_uuid = ajaxsyncSession::$values["AJAXSYNC_UUID"];
            ajaxsyncSession::session_write_close();
            $in_new = preg_replace("/$old_uuid/","$new_uuid",$in); // restart original request with new session id
            $out = self::ajaxsyncConnect($in_new);
            if (preg_match("/Session has expired|Session ID is valid|Failed:/",$out)) { // after a second failure, return error message
                return $out;
            }
            self::$answer = $out;
            return ""; // ok, answer is stored
        }
        else if (preg_match("/Failed:/",$out)) {
            return $out;
        }
        self::$answer = $out;
        return ""; // ok, answer is stored
    }

    // Interface between controller and AjaxSync server:

    /**
     * Using #request, starts a new Synchronizer session and stores the session ID
     * in the #ajaxsyncSession variable AJAXSYNC_UUID.
     * @return Error message, or nothing if the call was successful.
     */
    public static function startSession() {
        $in = AJAXSYNC_PASSWORD . "\nS\n";
        if ($errmsg = self::request($in)) {
            return $errmsg;
        }
        ajaxsyncSession::session_start();
        ajaxsyncSession::$values["AJAXSYNC_UUID"] = self::$answer;
        if ($listeners = ajaxsyncSession::$values["AJAXSYNC_LISTENERS"]) {
            unset(ajaxsyncSession::$values["AJAXSYNC_LISTENERS"]);
            foreach ($listeners as $key => $value) {
                if ($errmsg = self::listenerRegistration($key)) {
                    return $errmsg . ' (re-registration)';
                }
            }
        }
        ajaxsyncSession::session_write_close();
        return ""; // ok, session ID is stored
    }

    /**
     * Tests if a Synchronizer session is set in the #ajaxsyncSession variable
     * AJAXSYNC_UUID and checks if it is valid using #request.
     * @return Error message is the session is valid, otherwise nothing.
     */
    public static function testNoSession() {
        ajaxsyncSession::session_start();
        $uuid = ajaxsyncSession::$values["AJAXSYNC_UUID"];
        ajaxsyncSession::session_write_close();
        if (!$uuid) {
            return "";
        } // no ID set, ok
        $in = AJAXSYNC_PASSWORD . "\nT\n$uuid\n"; // test ID
        $errmsg = self::request($in);
        if (preg_match("/Session has expired/",$errmsg)) {
            // ok, the ID was NOT valid
            ajaxsyncSession::session_start();
            ajaxsyncSession::session_write_close();
            return "";
        }
        else if (preg_match("/Session ID is valid/",$errmsg)) {
            return "Failed: tried to open two sessions at once"; // error, the session IS valid
        }
        else {
            return $errmsg; // perhaps no connection to AjaxSync server?
        }
    }

    /**
     * Tests if a Synchronizer session is set in the #ajaxsyncSession variable
     * AJAXSYNC_UUID and if it is, tries to close it. If it was already invalid,
     * that is considered as success; if no other error occurs during the close
     * request, that is a success, too. In both cases the session variable
     * AJAXSYNC_UUID is unset.
     * @return Error message of the Synchronizer if not successful, otherwise nothing.
     */
    public static function closeSession() {
        ajaxsyncSession::session_start();
        $uuid = ajaxsyncSession::$values["AJAXSYNC_UUID"];
        ajaxsyncSession::session_write_close();
        if (!$uuid) {
            return "";
        } // already closed: ok ...
        $in = AJAXSYNC_PASSWORD . "\nC\n$uuid\n";
        $errmsg = self::request($in);
        if (preg_match("/Session has expired/",$errmsg)) {
            ajaxsyncSession::session_start();
            unset(ajaxsyncSession::$values["AJAXSYNC_UUID"]);
            ajaxsyncSession::session_write_close();
            return ""; // expired due to timeouts or never correct opened - ok too ...
        }
        else if ($errmsg) {
            return $errmsg; // perhaps no connection to AjaxSync server?
        }
        ajaxsyncSession::session_start();
        unset(ajaxsyncSession::$values["AJAXSYNC_UUID"]);
        ajaxsyncSession::session_write_close();
        return ""; // ok, session is closed
    }

    /**
     * If the listener was already registered, that is considered as success. If
     * not, try to register it using #requestReconnect. If that is successful,
     * update the #ajaxsyncSession listener array AJAXSYNC_LISTENERS.
     * @param $sig Register a listener for this signal.
     * @return Nothing if the listener was successfully registered, otherwise the
     * Synchronizer error message.
     */
    public static function listenerRegistration($sig) {
        ajaxsyncSession::session_start();
        $uuid = ajaxsyncSession::$values["AJAXSYNC_UUID"];
        if (ajaxsyncSession::$values["AJAXSYNC_LISTENERS"][$sig]) {
            return ""; // Double registration is allowed, it may happen f.e. during a second login try. Simply do nothing then.
        }
        ajaxsyncSession::session_write_close();
        $in = AJAXSYNC_PASSWORD . "\nL\n$uuid\n$sig\n";
        if ($errmsg = self::requestReconnect($in)) {
            return $errmsg;
        }
        ajaxsyncSession::session_start();
        ajaxsyncSession::$values["AJAXSYNC_LISTENERS"][$sig] = 1;
        ajaxsyncSession::session_write_close();
        return ""; // ok, listener is registered
    }

    /**
     * If the listener was not registered, return an error. Otherwise, try to remove
     * it using #requestReconnect. If that is successful, unset the entry in the
     * #ajaxsyncSession listener array AJAXSYNC_LISTENERS.
     * @param $sig Unregister the listener for this signal.
     * @return Nothing if the listener was successfully unregistered, otherwise the
     * Synchronizer error message.
     */
    public static function removeListener($sig) {
        ajaxsyncSession::session_start();
        $uuid = ajaxsyncSession::$values["AJAXSYNC_UUID"];
        if (!(ajaxsyncSession::$values["AJAXSYNC_LISTENERS"][$sig])) {
            return "no listener to $sig registered";
        }
        ajaxsyncSession::session_write_close();
        $in = AJAXSYNC_PASSWORD . "\nR\n$uuid\n$sig\n";
        if ($errmsg = self::requestReconnect($in)) {
            return $errmsg;
        }
        ajaxsyncSession::session_start();
        unset(ajaxsyncSession::$values["AJAXSYNC_LISTENERS"][$sig]);
        ajaxsyncSession::session_write_close();
        return ""; // ok, listener is removed
    }

    /**
     * Unregisters all listeners stored in the #ajaxsyncSession listener array
     * AJAXSYNC_LISTENERS using #removeListener.
     * @return Nothing if all listeners were successfully unregistered, otherwise the
     * Synchronizer error message.
     */
    public static function removeAllListeners() {
        ajaxsyncSession::session_start();
        if ($listeners = ajaxsyncSession::$values["AJAXSYNC_LISTENERS"]) {
            foreach ($listeners as $key => $value) {
                if ($errmsg = self::removeListener($key)) {
                    return $errmsg;
                }
            }
        }
        ajaxsyncSession::session_write_close();
    }

    /**
     * Sends a signal using #requestReconnect. Like in #request, the session ID is
     * taken from the #ajaxsyncSession variable AJAXSYNC_UUID.
     * @param $sig Send this signal.
     * @return Nothing if the signal was successfully sent, otherwise the Synchronizer
     * error message.
     */
    public static function kill($sig) {
        ajaxsyncSession::session_start();
        $uuid = ajaxsyncSession::$values["AJAXSYNC_UUID"];
        ajaxsyncSession::session_write_close();
        $in = AJAXSYNC_PASSWORD . "\nK\n$uuid\n$sig\n";
        if ($errmsg = self::requestReconnect($in)) {
            return $errmsg;
        }
        return ""; // ok, signal was sent
    }

    /**
     * Awaits a signal using #requestReconnect. Like in #request, the session ID is
     * taken from the #ajaxsyncSession variable AJAXSYNC_UUID.
     * @return Nothing if a signal was successfully received, otherwise the Synchronizer
     * error message. The returned signal itself is buffered in the private member
     * variable #$lastReceivedSignal.
     */
    public static function wait() {
        ajaxsyncSession::session_start();
        $uuid = ajaxsyncSession::$values["AJAXSYNC_UUID"];
        ajaxsyncSession::session_write_close();
        $in = AJAXSYNC_PASSWORD . "\nW\n$uuid\n";
        if ($errmsg = self::requestReconnect($in)) {
            return $errmsg;
        }
        if (self::$lastReceivedSignal = self::$answer) {
            return ""; // ok, received signal has been stored
        }
        else {
            return "no signal received";
        }
    }

    /**
     * This is simply a getter for #$lastReceivedSignal. However, in most cases the
     * returned signal is not really importand and will be ignored.
     * @return The last signal received with #wait.
     */
    public static function getReceivedSignal() {
        return self::$lastReceivedSignal;
    }
}
?>
