<?php
/****************************************************************************
 * Copyright (C) 2010  Michael Oberg
 * 
 * This file is part of the AjaxSync framework.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ****************************************************************************/

/**
 * This class handles the transport of lists through HTTP responses.
 * The elements may contain binary data.
 *
 * @author Michael Oberg
 */
final class ajaxsyncDataExchange {
    /**
     * A private constructor/clone does make creating instances of this class impossible.
     */
    private function __construct() { }

    /**
     * A private constructor/clone does make creating instances of this class impossible.
     */
    private function __clone() { }

    /**
     * This function replaces characters, which could collide with either HTTP or
     * AjaxSync control sequences, with their hexadecimal ASCII codes, escaped with
     * a leading percent character.
     * @param $s String to convert.
     * @return Escaped string.
     */
    private static function escape($s) {
        $s .= ""; // explicit conversion to string
        $len = strlen($s);
        $out = "";
        for($i = 0; $i < $len; $i++) {
            switch($s{$i}) {
                case chr(0):
                case "\n":
                case "\r":
                case "\t":
                case " ":
                case "&":
                case "=":
                case "%":
                    $out .= "%" . sprintf("%02x",ord($s{$i}));
                    break;
                default:
                    $out .= $s{$i};
            }
        }
        return $out;
    }

    /**
     * This function reverts what the function #escape has done, replacing the
     * escape sequences with the encoded characters.
     * @param $s String to convert.
     * @return Unescaped string.
     */
    private static function unescape($s) {
        $len = strlen($s);
        $i = 0;
        $out = "";
        while($i < $len) {
            if($s{$i} == '%') {
                $i++;
                $buf = $s{$i++};
                $buf .= $s{$i++};
                sscanf($buf,"%x",$c);
                $out .= chr($c);
            }
            else
                $out .= $s{$i++};
        }
        return $out;
    }

    /**
     * An AjaxSync variable is encoded in the following form:<br/>
     * &lt;<i>variable name</i>&gt; = &lt;<i>escaped value string</i>&gt;
     * @param $name Name of the variable.
     * @param $value Value of the variable. May contain binary data, but <b>not</b>
     * a compound data type (object, array etc.). Simple data types (strings,
     * numbers etc.) are allowed.
     * @return Encoded string. This is automatically a single line, because line
     * feeds contained in the variable value are replaced by the #escape function.
     */
    public static function writeVariable($name,$value) {
        return self::escape($name) . "=" . self::escape($value);
    }

    /**
     * A list of AjaxSync variables is encoded in the following form:<br/>
     * &lt;<i>variable name</i>&gt; = &lt;<i>escaped value string</i>&gt;<br/>
     * &lt;<i>variable name</i>&gt; = &lt;<i>escaped value string</i>&gt;<br/>
     * ...
     * @param $l Associative array of values. Simple data types are required
     * for these values (strings, numbers etc.).
     * @return Encoded string. Each value is encoded in a line.
     */
    public static function writeList($l) {
        $out = "";
        foreach($l as $key => $value) {
            //if (is_array($value) or is_buffer($value) or is_callable($value) or is_object($value) or is_resource($value))
            if (is_array($value) or is_callable($value) or is_object($value) or is_resource($value))
                throw(new Exception("Failed: Expecting string (getting: $value)"));
            $out .= ($out ? "&" : "") . self::writeVariable($key,$value);
        }
        return $out;
    }

    /**
     * Decodes an AjaxSync variable.
     * @param $s Encoded AjaxSync variable in the following form:<br/>
     * &lt;<i>variable name</i>&gt; = &lt;<i>escaped value string</i>&gt;
     * @return Array with two parameters: the name and the value of the AjaxSync
     * variable.
     */
    public static function readVariable($s) {
        list($n,$v) = explode("=",$s);
        return array(self::unescape($n),self::unescape($v));
    }

    /**
     * Decodes a list of AjaxSync variables.
     * @param $s List of encoded AjaxSync variables in the following form:<br/>
     * &lt;<i>variable name</i>&gt; = &lt;<i>escaped value string</i>&gt;<br/>
     * &lt;<i>variable name</i>&gt; = &lt;<i>escaped value string</i>&gt;<br/>
     * ...
     * @return Associative array of values. Key is the name of the variable.
     */
    public static function readList($s) {
        // Warning: explode with an empty or null element creates an array with
        // a single (empty) element
        if ($s) {
            $l = explode("&",$s);
            $out = array();
            foreach($l as $e) {
                list($n,$v) = self::readVariable($e);
                $out[$n] = $v;
            }
            return $out;
        }
        else {
            return array(); // empty array
        }
    }
}
?>
