<?php
/****************************************************************************
 * Copyright (C) 2010  Michael Oberg
 * 
 * This file is part of the AjaxSync framework.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ****************************************************************************/

/**
 * This class manages AjaxSync sessions. It is an equivalent to PHP sessions. But
 * on client side an AjaxSync session is stored in a persistent cookie with unlimited
 * lifetime, and on server side the informations attached to the session are stored
 * in the AjaxSync database.
 *
 * @author Michael Oberg
 */
final class ajaxsyncSession {
    /**
     * Contains the session variables.
     */
    public static $values = array();

    /**
     * Unique identifier for the session; stored in the clients cookie and in the database.
     */
    private static $sessionid;

    /**
     * Simple incremental counter / identifier for the session, only known to the server.
     */
    private static $sid;

    /**
     * Allow nested calls of session_start/session_write_close pairs; store the level here
     * (starting with 0).
     */
    private static $level = 0;

    /**
     * A private constructor/clone does make creating instances of this class impossible.
     */
    private function __construct() { }

    /**
     * A private constructor/clone does make creating instances of this class impossible.
     */
    private function __clone() { }

    /**
     * For a given #$sessionid, create a container for the session variables, if there
     * is none. Additionally, set a persistent cookie on the client.
     */
    private static function createDataContainer() {
        // try to set a cookie (valid till end of epoch - ca. 19.01.2038)
        if (!setcookie("AJAXSYNC_SESSION", self::$sessionid, 2147483647)) {
            throw new Exception('Failed: Could not set session cookie');
        }
        self::$sid = ajaxsyncDB::insertId("insert into sessioncontainer (session) values ('"
                . self::$sessionid . "')");
    }

    /**
     * For a given #$sessionid, get the session variables from the database container.
     * @return Success state (true/false)
     */
    private static function getDataContainer() {
        if ($ret = ajaxsyncDB::selectSingle("select content, id from sessioncontainer where session = '"
                . self::$sessionid . "'")) {
            self::$values = unserialize($ret["content"]);
            self::$sid = $ret["id"];
            return true;
        }
        // If nothing is returned from the select (return value is undef), then return false
        return false;
    }

    /**
     * Start an AjaxSync session by creating a session id if there is none, creating a cookie
     * on the client side if there is none, creating a data container on the server if there
     * is none, and getting the session data from the container if everything already exists.
     */
    public static function session_start() {
        if (self::$level++) {
            return; // do nothing except on level 0!
        }
        // first search for a session cookie with an AjaxSync session id
        if (self::$sessionid = $_COOKIE["AJAXSYNC_SESSION"]) {
            if (!self::getDataContainer()) {
                // no data container found - create a new empty one
                self::createDataContainer();
            }
        }
        else {
            // no session id set - create one
            self::$sessionid = md5(uniqid(mt_rand(), true));
            self::createDataContainer();
        }
    }

    /**
     * Close an AjaxSync session by putting the session variables in the data container.
     */
    public static function session_write_close() {
        if (--self::$level) {
            return; // do nothing except on level 0!
        }
        if (!isset(self::$sessionid)) {
            throw new Exception('Failed: AjaxSync session id is not set');
        }
        ajaxsyncDB::change("update sessioncontainer set content = '"
                . mysql_escape_string(serialize(self::$values))
                . "' where session = '" . self::$sessionid . "'");
    }

    /**
     * Destroying an AjaxSync session means deleting the data container, deleting all session
     * variables, deleting the cookie in the client and forgetting the session id.
     *
     * Be aware that you should clean up the database (queues etc.) before destroying the
     * session.
     */
    public static function session_destroy() {
        self::$values = array();
        ajaxsyncDB::change("delete from sessioncontainer where session = '" . self::$sessionid);
        if ($_COOKIE["AJAXSYNC_SESSION"]) {
            unset($_COOKIE["AJAXSYNC_SESSION"]);
            if (!headers_sent()) {
                // delete cookie on the client by deleting its value and setting its expire date in the past
                setcookie("AJAXSYNC_SESSION", '', time() - 3600);
            }
        }
        unset(self::$sessionid);
    }

    /**
     * Gets the internal (database) id for this session.
     */
    public static function getSID() {
        if (!self::$sid) {
            throw new Exception('Failed: Session not opened yet');
        }
        return self::$sid;
    }
}
?>
