<?php
/****************************************************************************
 * Copyright (C) 2010  Michael Oberg
 * 
 * This file is part of the AjaxSync framework.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ****************************************************************************/

/**
 * This class provides HTTP Digest Authentication against a single user.
 *
 * @author Michael Oberg
 */
class httpDigestAuth {
    private $realm = "";
    private $digest = "";

    /*
     * Creates the object and sets a pre-computed digest, f.e. from a
     * database, into the object. You have to make sure that the realm
     * matches the one used to compute the digest.
     *
     * Be aware that if you do not call that contructor with a digest,
     * you would have to set that value either by setDigest or
     * serverDigest, or the check functions would throw an exception.
     * @param $realm Realm
     * @param $digest Pre-computed digest
     */
    public function __construct($realm = "", $digest = "") {
        $this->realm = $realm;
        $this->digest = $digest;
    }

    /*
     * Sets a pre-computed digest, f.e. from a database, into the object.
     * You have to make sure that the realm matches the one used to
     * compute the digest.
     * @param $realm Realm
     * @param $digest Pre-computed digest
     */
    public function setDigest($realm, $digest) {
        $this->realm = $realm;
        $this->digest = $digest;
    }

    /*
     * Computes a digest from username, realm and password.
     * @param $user Username
     * @param $realm Realm
     * @param $pw Password
     * @returns Computed digest. This value and the digest are also stored
     * in the object.
     */
    public function serverDigest($user, $realm, $pw) {
        $this->realm = $realm;
        $this->digest = md5("$user:$realm:$pw");
        return $this->digest;
    }

    /*
     * Sends a HTTP 401 Not Authenticated Header and stops script execution.
     */
    private function sendHeader() {
        header('HTTP/1.1 401 Unauthorized');
        header('WWW-Authenticate: Digest realm="'
            . $this->realm
            . '",qop="auth",nonce="' 
            . uniqid()
            . '",opaque="'
            . md5($this->realm)
            . '"');
        die("Failed: Not authenticated");
    }

    /*
     * Parses the HTTP digest credentials string and creates an associative
     * array from it.
     * @returns An associative array with the digest encrypted credentials.
     */
    private function parseCredentials($txt) {
        $keys = array("nonce","nc","cnonce","qop","username","uri","response");
        foreach($keys as $i => $key) {
            preg_match_all("/(^|\\s|,)$key=(\"([^\"]*)\"|'([^']*)'|([^\\s,]+))/",$txt,$match);
            if ($match[5][0])
                $response[$key] = $match[5][0];
            else if ($match[4][0])
                $response[$key] = $match[4][0];
            else
                $response[$key] = $match[3][0];
        }
        return $response;
    }

    /*
     * Enforces and checks the HTTP Digest Authentication. Creates a 401
     * header and stops script execution if the credentials are invalid.
     */
    public function checkHttpAuth() {
        if (!($credentials = $_SERVER['PHP_AUTH_DIGEST']))
            $this->sendHeader(); // no digest credentials given: stop
        $this->checkDigest($this->parseCredentials($credentials)); // check; stop if that fails
    }

    /*
     * Checks digest authentication given by an associative array. Creates
     * a 401 header and stops script execution if the credentials are
     * invalid.
     */
    public function checkDigest($digestinfo) {
        if (!$this->digest)
            throw new Exception('Failed: tried to check digest credentials against undefined httpDigestAuth object');
        $methodUri = md5($_SERVER['REQUEST_METHOD'] . ":" . $digestinfo["uri"]);
        $valid = md5("{$this->digest}:{$digestinfo['nonce']}:{$digestinfo['nc']}:{$digestinfo['cnonce']}:{$digestinfo['qop']}:$methodUri");

        if ($digestinfo["response"] != $valid) {
            $this->sendHeader();
        }
    }
}
?>
