/*
 * Copyright (C) 2010  Michael Oberg
 *
 * This file is part of the AjaxSync framework.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/**
 * @fileoverview This is the AjaxSync JavaScript Browser Client.
 *
 * @author Michael Oberg
 */

/**
 * @class This class handles normal Ajax calls, but with AjaxSync failure recognition and
 * timeouts.
 * @param n Name of the object to be created, f.e. ajaxsync.sendConnection
 * @constructor
 **/
function ajaxsyncConnection(n) {
    /**
     * Name of the object ("this")
     * @variable
     */
    var objectName = n;
    /**
     * Private method. Creates an Ajax connection (browser independent).
     **/
    function getHTTPObject() {
        var xmlhttp;
        // browser hack: for IE Ajax is realised with ActiveX
        /*@cc_on
        @if(@_jscript_version >= 5)
            try {
                xmlhttp = new ActiveXObject("Msxml2.XMLHTTP"); // IE 6
            }
            catch(e) {
                try {
                    xmlhttp = new ActiveXObject("Microsoft.XMLHTTP"); // IE 5
                }
                catch(E) {
                    xmlhttp = false;
                }
            }
        @else
            xmlhttp=false;
        @end @*/
        if(!xmlhttp && typeof XMLHttpRequest != 'undefined') {
            try {
                xmlhttp = new XMLHttpRequest();
            }
            catch(e) {
                xmlhttp = false;
            }
        }
        return xmlhttp;
    }
    /**
     * Private member - Ajax connection object
     * @variable
     */
    var httpObject = getHTTPObject();
    /**
     * Private member - is a call running?
     * @variable
     */
    var running = false;
    /**
     * Private member - is a retrying call running?
     * @variable
     */
    var retryRunning = false;
    /**
     * Private member - was the call aborted?
     * @variable
     */
    var aborted = false;
    /**
     * Private member - Pointer to an error handler method, initiated via call
     * @param e Error message.
     * @function
     */
    var errorHandler = function(e) {};
    /**
     * Private member - Pointer to a success handler method, initiated via call
     * @param p Parameters returned by the called controller.
     * @function
     */
    var successHandler = function(p) {};
    /**
     * Private member - pointer to the timeout handler function, for stopping that handler
     * @variable
     */
    var timeoutHandler = false;
    /**
     * Private method - Ajax callback; calls the error / success handler in turn
     * @function
     */
    function ajaxCallback() {
        if (httpObject.readyState == 4) {
            window.clearTimeout(timeoutHandler);
            running = false;
            if (aborted) {
                aborted = false;
                // do nothing, especially don't call an error or success handler
                // (we do not know about the reason of the abort)
                return;
            }
            if (httpObject.status != 200) {
                errorHandler("Failed: got a HTTP "+httpObject.status+" response");
                if (httpObject.status == 0)
                    window.setTimeout("ajaxsync.reset();",100);
                return;
            }
            var respTxt = httpObject.responseText;
            // in Linux Konqueror, this can be null instead of an emtpy string, so ...
            if (!respTxt) {
                retryRunning = false; // success stops retries
                successHandler("");
                return;
            }
            // else:
            if ((respTxt.search (/Failed:/) != -1) || (respTxt.search (/Error [0-9]+:/) != -1)) {
                errorHandler(respTxt);
            }
            else {
                retryRunning = false; // success stops retries
                successHandler(respTxt);
            }
        }
    }
    /**
     * Ajax call.
     * @param c AjaxSync controller to be called - i.e. init, login, send, wait.
     * @param p POST parameters for the Ajax call.
     * @param t Timeout given in milliseconds.
     * @param e Pointer to an error handler. This has to be a static function, not
     * a member function.
     * @param s Pointer to a success handler. This has to be a static function, not
     * a member function.
     * @function
     */
    this.call = function(c,p,t,e,s) {
        if (running)
            return false;
        httpObject.open('POST', c, true);
        httpObject.setRequestHeader('Content-Type', 'application/x-www-form-urlencoded');
        httpObject.onreadystatechange = ajaxCallback;
        running = true;
        aborted = false;
        errorHandler = e;
        successHandler = s;
        httpObject.send(p);
        timeoutHandler = setTimeout(objectName+'.abort("Failed: Connection timed out."); ajaxsync.reset();',t);
        return true;
    };
    /**
     * Aborts a running Ajax call.
     * @param e Error message to be displayed with the error handler.
     * @function
     */
    this.abort = function(e) {
        if (!running)
            return;
        aborted = true; // don't call the error/success handler in the callback function
        httpObject.abort();
        running = false;
        retryRunning = false; // from now, the connection can be reused
        errorHandler(e); // instead, call the error handler directly
    };
    /**
     * Aborts a running Ajax call, but does not call the connections error handler.
     * @function
     */
    this.abortWithoutError = function() {
        if (!running)
            return;
        aborted = true; // don't call the error/success handler in the callback function
        httpObject.abort();
        running = false;
        retryRunning = false; // from now, the connection can be reused
    };
    /**
     * Ajax call to be automatically retried in case of an error.
     * @param c AjaxSync controller to be called - i.e. init, login, send, wait.
     * @param p POST parameters for the Ajax call.
     * @param t Timeout given in milliseconds.
     * @param e Pointer to an error handler. This has to be a static function, not
     * a member function.
     * @param s Pointer to a success handler. This has to be a static function, not
     * a member function.
     * @param r Time in milliseconds to wait till the next try.
     * @function
     */
    this.retryingCall = function(c,p,t,e,s,r) {
        function errorHandler(msg) {
            e(msg);
            setTimeout(objectName+'.callRetry();',r);
        }
        var This = this;
        this.callRetry = function() {
            This.call(c,p,t,errorHandler,s);
        }
        retryRunning = true;
        this.callRetry();
    };
    /**
     * Ajax call to be automatically retried in case of an error. But different from
     * retryingCall, this function handles parameters which can change during the
     * retries, and stops the retry loop if a condition is satisfied.
     * @param c AjaxSync controller to be called - i.e. init, login, send, wait.
     * @param p Static callback function without parameters which returns the POST
     * parameters for the Ajax call. The return value has to be non-empty, or the
     * call will not be made, and the retry loop stops.
     * @param con Static callback function without parameters which returns the loop
     * continuation condition (a value which can be converted into a boolean value).
     * The moment this function returns "false", the retry loop stops.
     * @param t Timeout given in milliseconds.
     * @param e Pointer to an error handler. This has to be a static function, not
     * a member function.
     * @param s Pointer to a success handler. This has to be a static function, not
     * a member function.
     * @param r Time in milliseconds to wait till the next try.
     * @function
     */
    this.retryingCallConditional = function(c,p,con,t,e,s,r) {
        function errorHandler(msg) {
            e(msg);
            if (con()) {
                setTimeout(objectName+'.callRetry();',r);
            }
        }
        var This = this;
        this.callRetry = function() {
            var str = p();
            if (str) {
                This.call(c,str,t,errorHandler,s);
            }
        }
        this.callRetry();
    };
}

/**
 * @namespace Pendant to the PHP ajaxsyncDataExchange class. This is a namespace for
 * the data exchange interface of AjaxSync, it enables the transport of associative
 * arrays and binary data between client and server.
 */
var ajaxsyncDataExchange = {
    /**
     * Escapes strings (even strings containing binary data) and other simple data types
     * for sending them to the server.
     * @param s String to be escaped.
     * @return The escaped string which will be sended to the server.
     * @function
     */
    escape: function(s) {
        var str = '' + s; // if s is a number, convert to string ...
        var a = str.split("");
        var i;
        var out = '';
        for(i = 0; i < a.length; i++) {
            switch(a[i]) {
                case String.fromCharCode(0):
                case "\n":
                case "\r":
                case "\t":
                case " ":
                case "&":
                case "=":
                case "%":
                    // JavaScript does not have sprintf, but the following construct does the same like:
                    // out += '%' + sprintf("%02x",s[i]).charCodeAt(0));
                    out += '%' + ('00' + (a[i]).charCodeAt(0).toString(16)).slice(-2);
                    break;
                default:
                    out += a[i];
            }
        }
        return out;
    },
    /**
     * Unescapes strings sended by the server for usage by the client.
     * @param s String to be unescaped.
     * @return The unescaped string.
     * @function
     */
    unescape: function(s) {
        if (!s)
            return "";
        var a = s.split("");
        var i = 0;
        var out = '';
        while(i < a.length) {
            if(a[i] == '%') {
                i++;
                var buf = '0x';
                buf += a[i++];
                buf += a[i++];
                out += String.fromCharCode(parseInt(buf));
            }
            else
                out += a[i++];
        }
        return out;
    },
    /**
     * Escapes a variable (name and value) which contains a simple data type
     * for sending to the server.
     * @param n Name of the variable.
     * @param v Value of the variable.
     * @return The escaped string which will be sended to the server.
     * @function
     */
    writeVariable: function(n,v) {
        return this.escape(n) + '=' + this.escape(v);
    },
    /**
     * Unescapes a variable returned by the server.
     * @param s String which contains the escaped variable.
     * @return An array with two components (0, 1) which contains the name (0)
     * and the value (1) of the variable.
     * @function
     */
    readVariable: function(s) {
        if (!s || (s.search(/=/) == -1))
            return Array();
        var a = s.split("=");
        a[0] = this.unescape(a[0]);
        a[1] = this.unescape(a[1]);
        return a;
    },
    /**
     * Escapes a list of variables (an associative array of simple data type
     * components) for sending to the server.
     * @param l The associative array to be escaped.
     * @return A string which contains the escaped list.
     * @function
     */
    writeList: function(l) {
        var out = '';
        for(var n in l)
            out += (out ? '&' : '') + this.writeVariable(n,l[n]);
        return out;
    },
    /**
     * Unescapes a list of variables encoded in a string.
     * @param s The string received from the server.
     * @return An associative array of simple data type components (a variable list).
     * @function
     */
    readList: function(s) {
        if (!s)
            return Object();
        var a = s.split("&");
        var out = Object(); // associative array
        for(var i = 0; i < a.length; i++) {
            var r = this.readVariable(a[i]);
            if (r.length)
                out[r[0]] = r[1];
        }
        return out;
    }
};

/**
 * @class This class is a container for event handlers.
 * @constructor
 */
function ajaxsyncRegistry() {
    /**
     * This array holds the names of all registered event handlers.
     * @variable
     */
    var handlerIdx = new Array();
    /**
     * This associative array holds pointers to the event handlers, indexed by name.
     * @variable
     */
    var handlerAss = new Object();
    /**
     * This empties the registry.
     * @function
     */
    this.clearHandlers = function() {
        handlerIdx = new Array();
        handlerAss = new Object();
    }
    /**
     * This adds one event handler to the registry.
     * @param f Handler name.
     * @param h Pointer to the handler function.
     * @function
     */
    this.addHandler = function(f,h) {
        handlerAss[f] = h;
        handlerIdx[handlerIdx.length] = f;
    }
    /**
     * This calls an event handler.
     * @param f Handler name.
     * @param p Parameter (string).
     * @function
     */
    this.callHandler = function(f,p) {
        if (handlerAss[f])
            (handlerAss[f])(p);
    }
    /**
     * This exports all handlers using ajaxsyncDataExchange for registering on the server.
     * @function
     */
    this.exportHandlers = function() {
        var handlerList = "";
        for (var i = 0; i < handlerIdx.length; i++) {
            handlerList += ajaxsyncDataExchange.writeVariable("fname", handlerIdx[i]) + "\n";
        }
        return handlerList;
    }
    /**
     * This array contains the registration status for each handler.
     * @variable
     */
    this.registered = new Array();
}

/**
 * @class This class is a container for pending calls to AjaxSync.
 * @param l Last call received by the server. This value comes from the server
 * during the initialization of AjaxSync.
 * @constructor
 */
function ajaxsyncCallBuffer(l) {
    /**
     * This array holds all pending calls not sent or not confirmed by the server.
     * @variable
     */
    var buffer = Object();
    /**
     * Index of the last call put in the buffer.
     * @variable
     */
    var lastCreatedCall = l;
    /**
     * Adds a call to the buffer.
     * @param f Function name.
     * @param p Parameter (string).
     * @param ts Server timeout in seconds.
     * @param tc Client timeout in seconds. Minimum timeout is 1 second.
     * @function
     */
    this.call = function(f,p,ts,tc) {
        var d = new Date();
        lastCreatedCall++;
        buffer[''+lastCreatedCall] = new Object();
        buffer[''+lastCreatedCall].fname = f;
        buffer[''+lastCreatedCall].param = p;
        buffer[''+lastCreatedCall].serverTimeout = ts;
        buffer[''+lastCreatedCall].clientTimeout = (tc > 1 ? tc : 1);
        buffer[''+lastCreatedCall].timestamp = d.getTime();
    }
    /**
     * Removes a call from the buffer. Called when the server confirms the call.
     * @param cid Client ID for this call.
     * @function
     */
    this.deleteCall = function(cid) {
        delete buffer[cid];
    }
    /**
     * Exports all calls using ajaxsyncDataExchange for sending them to the server.
     *
     * Calls whose client timeouts are exceeded will be deleted, not exported.
     * @returns Esscaped list of the calls.
     * @function
     */
    this.exportCalls = function() {
        var callList = "";
        for (id in buffer) {
            var d = new Date();
            if (buffer[id].timestamp + buffer[id].clientTimeout * 1000 < d.getTime()) {
                delete buffer[id];
            }
            else {
                var call = Object();
                call.cid = id;
                call.fname = buffer[id].fname;
                call.param = buffer[id].param;
                call.timeout = buffer[id].serverTimeout;
                callList += ajaxsyncDataExchange.writeList(call) + "\n";
            }
        }
        return callList;
    }
    /**
     * Count the calls in the buffer. Also automatically deletes all calls with client
     * timeout exceeded (without counting them).
     * @returns Count of the calls.
     * @function
     */
    this.count = function() {
        var c = 0;
        var d = new Date();
        for (id in buffer) {
            if (buffer[id].timestamp + buffer[id].clientTimeout * 1000 < d.getTime()) {
                delete buffer[id];
            }
            else {
                c++;
            }
        }
        return c;
    }
}

/**
 * @namespace Namespace for the AjaxSync JavaScript API
 */
var ajaxsync = {
    /**
     * For using ajaxsync.js in (perhaps cross-server) widgets, we need a possibility
     * to switch from relative to absolute paths. As long as the baseURL stays empty,
     * all paths are relative to the location of the document which loads the
     * ajaxsync.js.
     *
     * Do not forget the trailing slash.
     */
    baseURL: "",
    /**
     * Shows if the event loop (started by init, stopped by logout) is running or not.
     * @variable
     */
    eventLoopRunning: false,
    /**
     * After this amount of milliseconds the system confirms an event loop breakdown
     * and restarts the event loop automatically. Has to be longer than #connectionTimeout.
     * Default: 40 seconds.
     * @variable
     */
    waitTimeout: 40000,
    /**
     * After this amount of milliseconds a not responding connection will be closed.
     * Default: 30 seconds.
     * @variable
     */
    connectionTimeout: 30000,
    /**
     * After this amount of milliseconds a failed connection will be retried.
     * Default: 1 second.
     * @variable
     */
    retryDuration: 1000,
    /**
     * ID / number of the last call sent to the server.
     */
    lastSentCall: 0,
    /**
     * UNIX Timestamp in milliseconds of the last wait call. Used for monitoring of the
     * event loop; if wait isn't called after #waitTimeout milliseconds, the event loop
     * is restarted automatically.
     */
    waitLastCalled: 0,
    /**
     * ID / number of the last call received from the server.
     */
    lastReceivedCall: 0,
    /**
     * Most browsers will only allow two HTTP connections at the same time, therefore
     * two connection objects are created: sendConnection and waitConnection.
     * @variable
     */
    sendConnection: new ajaxsyncConnection("ajaxsync.sendConnection"),
    /**
     * Most browsers will only allow two HTTP connections at the same time, therefore
     * two connection objects are created: sendConnection and waitConnection.
     * @variable
     */
    waitConnection: new ajaxsyncConnection("ajaxsync.waitConnection"),
    registry: new ajaxsyncRegistry(),
    /**
     * Initialises the AjaxSync framework.
     * @param e Pointer to error handler.
     * @param l Pointer to login handler (shows the login dialog).
     * @function
     */
    init: function(e,l) {
        var s = function(msg) {
            if (msg.search(/OK: not logged in/) != -1)
                l(); // if not logged in, call login handler
            if (msg.search(/Last call was: ([0-9]+)/) != -1)
                ajaxsync.lastSentCall = parseInt(RegExp.$1);
            ajaxsync.pendingCalls = new ajaxsyncCallBuffer(ajaxsync.lastSentCall);
            ajaxsync.wait('');
        }
        this.loginHandler = l;
        this.errorHandler = e;
        this.eventLoopRunning = true;
        this.monitorEventLoop();
        this.sendConnection.retryingCall(this.baseURL+"init","",this.connectionTimeout,e,s,this.retryDuration);
    },
    /**
     * Logs in an AjaxSync user and tries to register all event handlers from the
     * AjaxSync registry. For each handler, a registration success state is stored
     * in the "registered" array of the registry.
     *
     * In case of invalid credentials call the login handler stored by "init".
     * @param u Username.
     * @param p Password.
     * @function
     */
    login: function(u,p) {
        var This = this;
        var s = function(msg) {
            if (msg.search(/OK: not logged in/) != -1)
                This.loginHandler(); // if not logged in, call login handler
            var msglist = msg.split("\n");
            This.registry.registered = new Object();
            for (var i = 0; i < msglist.length; i++) {
                var msgline = msglist[i];
                if (msgline && !(msgline.search(/OK: /) != -1)) {
                    var buffer = ajaxsyncDataExchange.readList(msgline)
                    for (j in buffer) {
                        This.registry.registered[j] = buffer[j];
                    }
                }
            }
        }
        this.sendConnection.retryingCall(
            this.baseURL+"login",
            "user="+escape(u)+"&password="+escape(p)+"&handler="+this.registry.exportHandlers(),
            this.connectionTimeout,
            this.errorHandler,
            s,
            this.retryDuration
            );
    },
    /**
     * Internal function. Callback for the "wait" Ajax call. In turn, it itself calls
     * "wait" in an endless loop (the AjaxSync event loop) and in doing so, it sends
     * confirmations for the events received from the server back.
     *
     * The events are processed using the event handler registry.
     * @param msg Message string received from the server.
     * @function
     */
    eventLoop: function(msg) {
        if (msg.search(/OK: not logged in/) != -1)
            ajaxsync.loginHandler(); // if not logged in, call login handler
        var msglist = msg.split("\n");
        if (msglist.length) {
            var conflist = new Object();
            for (var i = 0; i < msglist.length; i++) {
                var msgline = msglist[i];
                if (msgline && !(msgline.search(/OK: /) != -1)) {
                    var call = ajaxsyncDataExchange.readList(msgline);
                    var callid = parseInt(call.id);
                    if (isNaN(callid)) {
                        ajaxsync.errorHandler("Warning: Call id '"+call.id+"' is not a number - call to '"+call.name+"' ignored");
                    }
                    else {
                        if (callid > ajaxsync.lastReceivedCall) {
                            ajaxsync.lastReceivedCall = callid;
                            ajaxsync.registry.callHandler(call.name, call.param);
                        } // else: ignore already processed call
                        conflist[callid] = 1;
                    }
                }
            }
            ajaxsync.wait(ajaxsyncDataExchange.writeList(conflist));
        }
        else {
            ajaxsync.wait(""); // f.e. keepalive - no function call given
        }
    },
    /**
     * Waits for an event from the server.
     * @param confirmations Confirmations for the events received by the last
     * "wait" call.
     * @function
     */
    wait: function(confirmations) {
        this.waitConnection.retryingCall(this.baseURL+"wait",
            "ret="+escape(confirmations),
            this.connectionTimeout,
            this.errorHandler,
            this.eventLoop,
            this.retryDuration);
        var d = new Date();
        this.waitLastCalled = d.getTime();
    },
    /**
     * Sends a function call to the AjaxSync Server.
     * @param f Function name.
     * @param p Parameter (string).
     * @param ts Server timeout. This is the time the server tries to deliver the call
     * to the clients before giving up.
     * @param tc Client timeout. This is the time the client tries to deliver the call
     * to the AjaxSync server before giving up.
     * @function
     */
    send: function(f,p,ts,tc) {
        if (!(this.pendingCalls))
            return; // that means init was not run properly before - stop
        this.pendingCalls.call(f,p,ts,tc); // push call in the buffer
        var This = this;
        var parm = function() {
            return 'calls='+escape(This.pendingCalls.exportCalls());
        }
        var check = function() {
            return This.pendingCalls.count();
        }
        var s = function(msg) {
            if (msg.search(/OK: not logged in/) != -1) {
                This.loginHandler(); // if not logged in, call login handler
                msg = msg.replace(/OK: not logged in\n/g, "");
            }
            // returned confirmations; delete calls from buffer
            var confirmations = ajaxsyncDataExchange.readList(msg);
            for (var id in confirmations) {
                This.pendingCalls.deleteCall(id);
            }
            if (This.pendingCalls.count()) {
                // still un-confirmed calls in the buffer (perhaps send was called
                // during the servers compute time) - push them to the server
                // immediately
                This.sendConnection.retryingCallConditional(
                    This.baseURL+"send",
                    parm,
                    check,
                    This.connectionTimeout,
                    This.errorHandler,
                    s,
                    This.retryDuration
                    );
            }
        }
        this.sendConnection.retryingCallConditional(
            this.baseURL+"send",
            parm,
            check,
            this.connectionTimeout,
            this.errorHandler,
            s,
            this.retryDuration
            );
    },
    /**
     * Logout and stop event loop. This does not work like a (re)login, which is basically
     * only a re-registration of the listeners in context of the given username, but the
     * logout function re-initializes all objects except the ajaxsync object itself.
     * Re-usage enforces another call to the #init function.
     * @function
     */
    logout: function() {
        this.eventLoopRunning = false;
        this.sendConnection.abortWithoutError();
        this.waitConnection.abortWithoutError();
        this.sendConnection = new ajaxsyncConnection("ajaxsync.sendConnection");
        this.waitConnection = new ajaxsyncConnection("ajaxsync.waitConnection");
        this.registry = new ajaxsyncRegistry();
    },
    /**
     * Stop event loop, like in logout. Then restart event loop. Does not
     * use init; instead, on server side the controller initReset is called,
     * which does not test for an already opened session.
     * @function
     */
    reset: function() {
        this.errorHandler('Failed: Resetting event loop');
        this.waitConnection.abortWithoutError();
        this.waitConnection = new ajaxsyncConnection("ajaxsync.waitConnection");
        var s = function(msg) {
            if (msg.search(/OK: not logged in/) != -1)
                ajaxsync.loginHandler(); // if not logged in, call login handler
            if (msg.search(/Last call was: ([0-9]+)/) != -1)
                ajaxsync.lastSentCall = parseInt(RegExp.$1);
            ajaxsync.pendingCalls = new ajaxsyncCallBuffer(ajaxsync.lastSentCall);
            ajaxsync.wait('');
        }
        this.sendConnection.retryingCall(this.baseURL+"initReset","",this.connectionTimeout,this.errorHandler,s,this.retryDuration);
    },
    /**
     * This function is a background loop which checks the event loop every second. If the
     * event loop breaks down, this function will restart it.
     * @function
     */
    monitorEventLoop: function() {
        if (this.eventLoopRunning) {
            var d = new Date();
            var t = d.getTime();
            if (this.waitLastCalled && (t - this.waitLastCalled >= this.waitTimeout))
                this.reset();
            window.setTimeout("ajaxsync.monitorEventLoop();",1000);
        }
    }
};
